import datetime
import functools
import logging
import signal

_logger = logging.getLogger(__name__)


def log_call_time(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        current_time = datetime.datetime.now()
        _logger.info(f"Call: {func.__name__}.")
        result = func(*args, **kwargs)
        delta = datetime.datetime.now() - current_time

        _logger.info(f"Call: {func.__name__} took {delta.total_seconds()}s.")
        return result

    return wrapper


class GracefulStopper:
    stop_now = False
    stop_now_requests = 0

    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, *_):
        self.stop_now = True
        self.stop_now_requests += 1

        if self.stop_now_requests > 2:
            _logger.info("Received 3 stop signals, stopping NOW!")
            exit(1)



from io import StringIO


def dict_to_yaml_string(data):
    """
    Converts a dictionary to a YAML string, ensuring multiline strings use the '|' block style.

    Args:
        data: The input data dictionary.

    Returns:
        A string containing the YAML representation of the input data.
    """
    try:
        from ruamel.yaml import YAML
        from ruamel.yaml.scalarstring import LiteralScalarString

    except ImportError:
        raise ImportError(
            "Please install the 'ruamel.yaml' package to use this function."
        )

    def process_multiline_strings(data):
        """
        Recursively processes the input data, wrapping multiline strings in LiteralScalarString.

        Args:
            data: The input data (dict, list, or string).

        Returns:
            The processed data with multiline strings wrapped.
        """

        if isinstance(data, dict):
            return {k: process_multiline_strings(v) for k, v in data.items()}
        elif isinstance(data, str):
            if "\n" in data:
                return LiteralScalarString(data)
            else:
                return data
        elif isinstance(data, list):
            return [process_multiline_strings(v) for v in data]
        else:
            return data

    yaml = YAML()
    yaml.default_flow_style = False
    yaml.allow_unicode = True

    processed_data = process_multiline_strings(data)

    stream = StringIO()
    yaml.dump(processed_data, stream)
    return stream.getvalue()



from collections import deque


class SimpleCache:
    def __init__(self, max_size=100):
        self.max_size = max_size
        self.cache = {}
        self.queue = deque()

    def put(self, key, data):
        if key in self.cache:
            self.cache[key] = data
            self.queue.remove(key)
            self.queue.append(key)
            return
        if len(self.cache) >= self.max_size:
            old_key = self.queue.popleft()
            del self.cache[old_key]
        self.cache[key] = data
        self.queue.append(key)

    def get(self, key):
        return self.cache.get(key, None)



from enum import Enum, EnumType


def _enum_str(self):
    return f"{self.__class__.__name__}.{self.name}"


def _enum_str_attr(self):
    # format an enum such that it is valid python code
    additional_attrs = []
    for k, v in vars(self).items():
        if not k.startswith("_"):
            if isinstance(v, str):
                additional_attrs.append((k, f"'{v}'"))
            else:
                additional_attrs.append((k, v))
    if len(additional_attrs) > 0:
        return f"{self.__class__.__name__}.{self.name}({','.join([f'{k}={v}' for k,v in additional_attrs])})"
    return f"{self.__class__.__name__}.{self.name}"


def _enum_repr(self):
    return _enum_str(self)


def add_enum_repr(enum_class):
    enum_class.__str__ = _enum_str
    enum_class.__repr__ = _enum_repr
    return enum_class


def add_enum_repr_attr(enum_class):
    enum_class.__str__ = _enum_str_attr
    enum_class.__repr__ = _enum_str
    return enum_class


def add_enum_attrs(attr_dict):
    attr_set = None
    enum_classes = set([k.__class__ for k in attr_dict.keys()])
    if len(enum_classes) > 1:
        raise ValueError("All enums must be of the same class")

    enum_class = enum_classes.pop()
    members = set([e for e in enum_class])

    enum_names_keys = set(attr_dict.keys())

    if members != enum_names_keys:
        missing_names = [e.name for e in members - enum_names_keys]
        raise ValueError(
            f"All enums must be defined in the enum class, but {missing_names} are missing."
        )

    for enum_key, enum_attrs in attr_dict.items():
        if attr_set is None:
            attr_set = set(enum_attrs.keys())
        else:
            if set(enum_attrs.keys()) != attr_set:

                missing = attr_set - set(enum_attrs.keys())
                too_much = set(enum_attrs.keys()) - attr_set

                raise ValueError(
                    f"All enums must have the same attributes. {enum_key} does not have the same attributes as others, missing: {missing}, new: {too_much}"
                )
        for attr_key, value in enum_attrs.items():
            setattr(enum_key, attr_key, value)



import os

import pkg_resources


def _calc_package_and_path(sub_package_name, filename):
    root_package_name = sub_package_name.split(".")[0]
    relative_path = os.path.join(*sub_package_name.split(".")[1:], filename)
    return root_package_name, relative_path


def get_resource_path(sub_package_name, filename):
    root_package_name, relative_path = _calc_package_and_path(
        sub_package_name, filename
    )
    return pkg_resources.resource_filename(root_package_name, relative_path)


def get_resource_string(sub_package_name, filename):
    root_package_name, relative_path = _calc_package_and_path(
        sub_package_name, filename
    )
    return pkg_resources.resource_string(root_package_name, relative_path).decode(
        "utf-8"
    )



import json
import logging
import time

import requests
from box import Box
from poemai_utils.openai.openai_model import OPENAI_MODEL

_logger = logging.getLogger(__name__)


class PydanticLikeBox(Box):
    def dict(self):
        return self.to_dict()


class AskLean:
    OPENAI_MODEL = (
        OPENAI_MODEL  # to make it easier to import / access, just use Ask.OPENAI_MODEL
    )

    def __init__(
        self,
        openai_api_key,
        model="gpt-4",
        base_url="https://api.openai.com/v1/chat/completions",
        timeout=60,
        max_retries=3,
        base_delay=1.0,  # seconds
    ):
        self.openai_api_key = openai_api_key
        self.model = model
        self.base_url = base_url
        self.timeout = timeout
        self.max_retries = max_retries
        self.base_delay = base_delay

    def ask(
        self,
        messages,
        model=None,
        temperature=0,
        max_tokens=600,
        stop=None,
        tools=None,
        tool_choice=None,
        json_mode=False,  # still just a placeholder
        response_format=None,
        additional_args=None,
    ):
        use_model = model if model is not None else self.model

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.openai_api_key}",
        }

        data = {"model": use_model, "messages": messages, "temperature": temperature}

        if max_tokens is not None:
            data["max_tokens"] = max_tokens

        if stop is not None:
            data["stop"] = stop

        if tools is not None:
            data["tools"] = tools
        if tool_choice is not None:
            data["tool_choice"] = tool_choice

        # Add response_format if provided
        if response_format is not None:
            data["response_format"] = response_format

        if additional_args is not None:
            data.update(additional_args)

        for attempt in range(self.max_retries):
            try:
                _logger.debug(
                    f"Sending request to OpenAI API: url={self.base_url} data={data}"
                )
                response = requests.post(
                    self.base_url,
                    headers=headers,
                    data=json.dumps(data),
                    timeout=self.timeout,
                )

                if response.status_code == 200:
                    response_json = response.json()
                    _logger.debug(f"Received response from OpenAI API: {response_json}")
                    retval = PydanticLikeBox(response_json)
                    return retval

                else:
                    # Non-200 response. Retry if it's a server error.
                    if (
                        500 <= response.status_code < 600
                        and attempt < self.max_retries - 1
                    ):
                        sleep_time = self.base_delay * (2**attempt)
                        time.sleep(sleep_time)
                        continue
                    else:
                        # Non-retryable error or last attempt
                        raise RuntimeError(
                            f"OpenAI API call failed with status {response.status_code}: {response.text}"
                        )
            except requests.exceptions.RequestException as e:
                # Network or connection error - retry if possible
                if attempt < self.max_retries - 1:
                    sleep_time = self.base_delay * (2**attempt)
                    time.sleep(sleep_time)
                else:
                    raise RuntimeError(f"OpenAI API request failed: {e}")

        # If we got here, it means we exhausted all retries
        raise RuntimeError("Failed to get a successful response after all retries.")
